﻿#region License Information
/* HeuristicLab
 * Copyright (C) Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HEAL.Attic;

namespace HeuristicLab.Algorithms.DataAnalysis {
  [StorableType("709D0C3B-EFDC-4113-B009-E0CE95A835F6")]
  // base class for GaussianProcessModelCreators (specific for classification and regression)
  public abstract class GaussianProcessModelCreator : SingleSuccessorOperator {
    private const string HyperparameterParameterName = "Hyperparameter";
    private const string MeanFunctionParameterName = "MeanFunction";
    private const string CovarianceFunctionParameterName = "CovarianceFunction";
    private const string ModelParameterName = "Model";
    private const string NegativeLogLikelihoodParameterName = "NegativeLogLikelihood";
    private const string NegativeLogPseudoLikelihoodParameterName = "NegativeLogPseudoLikelihood (LOOCV)";
    private const string HyperparameterGradientsParameterName = "HyperparameterGradients";
    protected const string ScaleInputValuesParameterName = "ScaleInputValues";

    #region Parameter Properties
    // in
    public ILookupParameter<RealVector> HyperparameterParameter {
      get { return (ILookupParameter<RealVector>)Parameters[HyperparameterParameterName]; }
    }
    public ILookupParameter<IMeanFunction> MeanFunctionParameter {
      get { return (ILookupParameter<IMeanFunction>)Parameters[MeanFunctionParameterName]; }
    }
    public ILookupParameter<ICovarianceFunction> CovarianceFunctionParameter {
      get { return (ILookupParameter<ICovarianceFunction>)Parameters[CovarianceFunctionParameterName]; }
    }
    // out
    public ILookupParameter<IGaussianProcessModel> ModelParameter {
      get { return (ILookupParameter<IGaussianProcessModel>)Parameters[ModelParameterName]; }
    }
    public ILookupParameter<RealVector> HyperparameterGradientsParameter {
      get { return (ILookupParameter<RealVector>)Parameters[HyperparameterGradientsParameterName]; }
    }
    public ILookupParameter<DoubleValue> NegativeLogLikelihoodParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters[NegativeLogLikelihoodParameterName]; }
    }
    public ILookupParameter<DoubleValue> NegativeLogPseudoLikelihoodParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters[NegativeLogPseudoLikelihoodParameterName]; }
    }
    public ILookupParameter<BoolValue> ScaleInputValuesParameter {
      get { return (ILookupParameter<BoolValue>)Parameters[ScaleInputValuesParameterName]; }
    }
    #endregion

    #region Properties
    protected RealVector Hyperparameter { get { return HyperparameterParameter.ActualValue; } }
    protected IMeanFunction MeanFunction { get { return MeanFunctionParameter.ActualValue; } }
    protected ICovarianceFunction CovarianceFunction { get { return CovarianceFunctionParameter.ActualValue; } }
    public bool ScaleInputValues { get { return ScaleInputValuesParameter.ActualValue.Value; } }
    #endregion

    [StorableConstructor]
    protected GaussianProcessModelCreator(StorableConstructorFlag _) : base(_) { }
    protected GaussianProcessModelCreator(GaussianProcessModelCreator original, Cloner cloner) : base(original, cloner) { }
    protected GaussianProcessModelCreator()
      : base() {
      // in
      Parameters.Add(new LookupParameter<RealVector>(HyperparameterParameterName, "The hyperparameters for the Gaussian process model."));
      Parameters.Add(new LookupParameter<IMeanFunction>(MeanFunctionParameterName, "The mean function for the Gaussian process model."));
      Parameters.Add(new LookupParameter<ICovarianceFunction>(CovarianceFunctionParameterName, "The covariance function for the Gaussian process model."));
      // out
      Parameters.Add(new LookupParameter<IGaussianProcessModel>(ModelParameterName, "The resulting Gaussian process model"));
      Parameters.Add(new LookupParameter<RealVector>(HyperparameterGradientsParameterName, "The gradients of the hyperparameters for the produced Gaussian process model (necessary for hyperparameter optimization)"));
      Parameters.Add(new LookupParameter<DoubleValue>(NegativeLogLikelihoodParameterName, "The negative log-likelihood of the produced Gaussian process model given the data."));
      Parameters.Add(new LookupParameter<DoubleValue>(NegativeLogPseudoLikelihoodParameterName, "The leave-one-out-cross-validation negative log pseudo-likelihood of the produced Gaussian process model given the data."));


      Parameters.Add(new LookupParameter<BoolValue>(ScaleInputValuesParameterName,
        "Determines if the input variable values are scaled to the range [0..1] for training."));
      Parameters[ScaleInputValuesParameterName].Hidden = true;
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (!Parameters.ContainsKey(ScaleInputValuesParameterName)) {
        Parameters.Add(new LookupParameter<BoolValue>(ScaleInputValuesParameterName,
          "Determines if the input variable values are scaled to the range [0..1] for training."));
        Parameters[ScaleInputValuesParameterName].Hidden = true;
      }
      if (!Parameters.ContainsKey(NegativeLogPseudoLikelihoodParameterName)) {
        Parameters.Add(new LookupParameter<DoubleValue>(NegativeLogPseudoLikelihoodParameterName,
          "The leave-one-out-cross-validation negative log pseudo-likelihood of the produced Gaussian process model given the data."));
      }
    }
  }
}
